// to do:
// X 1) update options input
// X 1a) add PRECISION MAX def and base options limits on it
// X 2) revise and add outputs corresponding to new options  1
// X 3) convert outputs to SI units  1
// X 4) update Terry's data files   1/2
//   5) rerun Terry's data to verify no errors added 2
//   6) delete old Terry's data and data files 1/4
//
/***********************************************************************
   This program prompts the user for coil geometry and room size and
   then calculates the electric field in the room from this data.
   From this data, the coil+terminal C and resonant frequency are
   calculated.

   This version adds units to dimensions.  Allowed units are in, ft,
   mm, cm, and m.  The "base" unit for non-discrete distances is
   meters.


   Creation date: 1/26/1999 - 1/2?/1999 (based on existing program)
***********************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h> /* needed for uppercpy */
#include <time.h>

/* Add debugging code?   Comment out line below if not. */
//#define DEBUG_freq

/* Default profile file if none specified on command line */
#define FILENAME_PROF_DEFAULT "fr_prof.txt"

/* Number of char used printing exponent (Implementation dependent) */
#define EXPONENT_WIDTH 3

/* Number of grids used by this version of the program */
#define N_GRID 1

/* Maximum precision used in output routines */
#define PRECISION_MAX 10
#define PRECISION_MAX_STRING "10"

/* Elements required for a SI-prefix conversion string:
      + 1 for "-" and leading space in certain
        cases
      PRECISION_MAX for digits
      +1 for '.'
      +1 for 'e'
      +1 for '+/-' on exponent
      +EXPONENT_WIDTH for characters making up exponent
      +1 for ' ' at end for adding a space after number if
        conversion is not possible due to size of the number
      +1 for '\0' to terminate the string */
#define SI_STRING_SIZE PRECISION_MAX+EXPONENT_WIDTH+6

/* Data type used to store guard matrix.  Reasonable possibilities
   would be char, unsigned char, short, unsigned short, int,
   unsigned int.  The preferred method would be the one that runs
   fastest (assuming not limited by memory size).  Bits would be ideal
   in terms of storage requirements, but are not fully portable. */
#define GUARD_TYPE int

/* some constants */
#define PI 3.14159265358979323846264338327950288419716939937510
#define M_PER_FOOT 0.3048 /* meters per foot conversion */
#define M_PER_INCH 0.0254 /* meters per inch conversion */
#define E0 8.854187818E-12 /* permittivity of free space F/m */
#define MIN_ANGLE_DIF 0.00001 /* for arc inputs */
#define N_ITER_MAX_DEFAULT 1000 /* default max number of interations */
#define RESIDUAL_MAX_DEFAULT 0.001 /* default maximum residual */

/* constant to input from data files */
#define INPUT_STRING_MAX 200
#if(FILENAME_MAX>INPUT_STRING_MAX) /* may input a file name */
#undef INPUT_STRING_MAX
#define INPUT_STRING_MAX FILENAME_MAX
#endif


/* error codes */
#define E_BAD_COMMAND_LINE 1
#define E_BAD_INFILE 2
#define E_BAD_OUTFILE 3
#define E_BAD_DATA 4
#define E_NO_MEMORY 5
#define E_GRID_ALLOC 6
#define E_DIVIDE_BY_0 7

/* macro to handle error in main() */
#define FATAL_ERROR(error_code,error_message)\
  do{\
    (void)fprintf(stderr,\
        "   *** %s\n"\
        "   *** Program is stopped.\n",error_message);\
    xrc = error_code;\
    goto exit;\
    }while(0)

/* macro to handle error in function other than main() */
#define FUNCTION_ERROR(error_code,error_message)\
  do{\
    (void)fprintf(stderr,\
        "   *** %s\n",error_message);\
    return error_code;\
    }while(0)

#define MAX(a,b) ((a)>(b)? (a) : (b))
#define MIN(a,b) ((a)<(b)? (a) : (b))
#define SQ(a) ((a)*(a))

/* options */
struct Option{
  int n_grid_point; /* number of grid points used */
  int match_grid_to_sec; /* Should room dimensions be adjusted for 
                     secondary to coincide with a line on the grid? */
  int draw_sec_ground; /* Should a secondary ground be added from the
                          secondary to the floor? */
  float residual_max; /* maximum residual for iterations */
  int n_iter_max; /* maximum number of iterations */

  /* options related to output */
  int print_room; /* Should room data be printed in output file?
                                         If so, how much precision? */
  int print_coil; /* Should coil data be printed in output file?
                                         If so, how much precision? */
  int print_guard; /* Should guard data be printed in output file? */
  int print_range; /* Should range data be printed in output file? */
  int print_volt; /* Should voltage data be printed in output file?
                                         If so, how much precision? */
  int print_cap; /* Should cap data be printed in output file?
                                         If so, how much precision? */
  int print_freq; /* Should frequency data be printed in output file?
                                         If so, how much precision? */
  };

/* room definition */
struct Room{
  float r; /* radius of "cylinder room" */
  float z; /* height of "cylinder room" */
  int n_r; /* number of points in radial direction */
  int n_z; /* number of points vertically */
  float dl; /* step size */
  };

/* coil definition */
struct Coil{
  float pri_h_i; /* height of inner (lowermost) turn of primary */
  float pri_h_o; /* height of outer (uppermost) turn of primary */
  float pri_r_i; /* radius of inner (lowermost) turn of primary */
  float pri_r_o; /* radius of outer (uppermost) turn of primary */
  float sec_h_b; /* height of base of secondary */
  float sec_h_t; /* height of top of secondary */
  float sec_r; /* radius of secondary */
  float sec_l; /* secondary inductance */
  float term_h; /* terminal height */
    /* integer versions based on descritized model */
  int n_pri_h_i; /* height of inner (lowermost) turn of primary */
  int n_pri_h_o; /* height of outer (uppermost) turn of primary */
  int n_pri_r_i; /* radius of inner (lowermost) turn of primary */
  int n_pri_r_o; /* radius of outer (uppermost) turn of primary */
  int n_sec_h_b; /* height of base of secondary */
  int n_sec_h_t; /* height of top of secondary */
  int n_sec_r; /* radius of secondary */
  int n_term_h; /* terminal height */
  };

/* structure to store range of grid points */
struct Range{
  int low; /* low end of range of points */
  int high; /* high end of range of points */
  };

/* structure to store composite grid data */
struct Grid_data{
  struct Room room; /* include the room since it has extents of
                       the grid */
  float *v; /* voltage array */
  float **v_row; /* pointes to rows in voltage array */
  GUARD_TYPE *border_guard; /* border guard array */
  GUARD_TYPE **border_guard_row; /* border guard array row */
  struct Range *range_h; /* horizontal ranges not guarded in
                            vertical rows */
  struct Range **range_h_row; /* pointers to rows of ranges */
  int *range_h_count; /* number of horizontal ranges in a vert row */
  };

  /* Structure to store system parameters.  All elements are pointers
     to other places where the data is actually stored */
  struct System_parm{
    /* files */
    char *filename_opt; /* option filename */
    char *filename_room; /* room filename */
    char *filename_pri; /* primary coil filename */
    char *filename_sec; /* secondary coil filename */
    char *filename_term; /* terminal filename */
    char *filename_out; /* output filename */
    char *filename_diag; /* diagnostic filename */
    /* component heights */
    float *pri_h_i; /* height of primary coil */
    float *sec_h_b; /* height of secondary coil */
    float *term_h; /* height of terminal */
    };

/* Memory allocation functions */
int allocate_voltage_memory(struct Grid_data *grid);
int allocate_grid_memory(struct Grid_data *grid);

/* functions to input data */
int input_profile(FILE *fp_prof, struct System_parm *system_parm);
int input_option(FILE *fp_in, struct Option *option);
int input_room(FILE *fp_room, struct Room *room);
int input_coil(FILE *fp_pri, FILE *fp_sec, struct Room *room,
    struct Coil *coil);
int input_terminal(FILE *fp_term, struct Grid_data *grid,
    struct Coil *coil);
int input_string(FILE *fp_in, char *instring);
int input_string_echo(FILE *fp_in, char *prompt, char *instring);
int input_float_echo_range(FILE *fp_in, char *prompt, float *x,
    float low, float high);
int input_float_echo(FILE *fp_in, char *prompt, float *x);
int input_int_echo_range(FILE *fp_in, char *prompt, int *x,
    int low, int high);
int input_int_echo(FILE *fp_in, char *prompt, int *x);
int input_length_echo(FILE *fp_in, char *prompt, float *x);
int input_length_echo_range(FILE *fp_in, char *prompt, float *x,
    float low, float high);


/* functions to check ranges of numbers */
int check_range_float(float x, float low, float high);
int check_range_int(int x, int low, int high);


/* function to perform iteration in solving Laplace's equation */
void laplace_iter_interval_c(struct Grid_data *grid,
    float residual_max, int* converge);

/* Function uppercpy is identical to strcpy except that it uppercases
   any lower case letters in src while performing the copy */
char *uppercpy(
    char *dst, /* destination */
    char *src); /* pointer to source string */

/* Function si_prefix converts a number to string consisting of a
   number + SI prefix -- e.g. 30000 -> "30 k" */
char *si_prefix(double x, int precision, char *work_string);

/* Function find conversion from length unit to meters */
int length_unit_factor(const char *length_unit_string,float *factor);


/* Functions to perform output */
void print_room(FILE* fp, struct Grid_data *grid, int precision);
void print_coil(FILE *fp, struct Coil *coil, int precision);
void print_guard(FILE *fp, GUARD_TYPE **a, int r, int c);
void print_range(FILE* fp, struct Grid_data *grid, int precision);
void print_volt(FILE *fp, float **a, int r, int c, int precision);


#ifdef DEBUG_freq
#define TOL_V 0.001
int compare_v(char* filename, char* tag, struct Grid_data *grid1,
    struct Grid_data *grid2);
#endif


int main(int argc, char *argv[]);






/* main program */
int main(int argc, char *argv[]){

  struct Option option; /* options used in run */
  struct Coil coil; /* coil definition */
  struct Grid_data grid[N_GRID]; /* grid description */

  int xrc=0; /* program exit return code */

  /* File pointers */
  FILE* fp_prof=(FILE*)NULL; /* profile file */
  FILE* fp_opt=(FILE*)NULL; /* input file for options */
  FILE* fp_room=(FILE*)NULL; /* input file for room data */
  FILE* fp_pri=(FILE*)NULL; /* input file for primary coil data */
  FILE* fp_sec=(FILE*)NULL; /* input file for secondary coil data */
  FILE* fp_term=(FILE*)NULL; /* input file for terminal data */
  FILE* fp_out=(FILE*)NULL; /* output file */
  FILE* fp_diag=(FILE*)NULL; /* diagnostic file */

  /* File names */
  char filename_prof_default[]=FILENAME_PROF_DEFAULT;
                                   /* default profile filename */
  char *filename_prof;
  char filename_opt[FILENAME_MAX+1]; /* option filename */
  char filename_room[FILENAME_MAX+1]; /* room filename */
  char filename_pri[FILENAME_MAX+1]; /* primary coil filename */
  char filename_sec[FILENAME_MAX+1]; /* secondary coil filename */
  char filename_term[FILENAME_MAX+1]; /* terminal filename */
  char filename_out[FILENAME_MAX+1]; /* output filename */
  char filename_diag[FILENAME_MAX+1]; /* diagnostic filename */

  /* Initialize the elements to the actual storage locations */
  struct System_parm system_parm={
    filename_opt, /* option filename */
    filename_room, /* room filename */
    filename_pri, /* primary coil filename */
    filename_sec, /* secondary coil filename */
    filename_term, /* terminal filename */
    filename_out, /* output filename */
    filename_diag, /* diagnostic filename */
    /* component heights */
    &coil.pri_h_i, /* height of primary coil */
    &coil.sec_h_b, /* height of primary coil */
    &coil.term_h, /* height of primary coil */
    };

  time_t run_time=time((time_t*)NULL); /* time of run */

  
  /* Initialize all grid pointers to nulls to assist with cleanup */
  {
    int i;
    for(i=0;i<N_GRID;i++){
      grid[i].v=(float*)NULL;
      grid[i].v_row=(float**)NULL;
      grid[i].border_guard=(GUARD_TYPE*)NULL;
      grid[i].border_guard_row=(GUARD_TYPE**)NULL;
      grid[i].range_h=(struct Range*)NULL;
      grid[i].range_h_row=(struct Range**)NULL;
      grid[i].range_h_count=(int*)NULL;
      }
    } /* end of block initializing grid pointers */

  
  (void)fprintf(stdout,
      "This program calculates a Tesla Coil's resonant frequency.\n"
      "Lenghts may be suffixed by mm, cm, m, in, or ft.\n"
      "The default unit of length is meters.\n");



  /* Parse command line for the profile file.  
     A valid command line consists of the name of the program
     optionally followed by a profile file that has data on where to
     get other information.

     If no argument or too many, print some help information in case
     the user is confused. */
  if(argc>2){
    (void)fprintf(stdout,
        "Format:  %s profile_filename\n"
        "profile_filename defaults to %s if none is supplied\n",
        argv[0],FILENAME_PROF_DEFAULT);
    FATAL_ERROR(E_BAD_COMMAND_LINE,
        "Command line was not specified properly");
    }

  if(argc==1) filename_prof=filename_prof_default;
  else filename_prof=argv[1];

  /* Open profile file if possible. */
  if(filename_prof[0]=='-'&&filename_prof[1]=='\0'){ /* stdin */
    (void)fprintf(stdout,
        "Data will be input from stdin.\n");
    fp_prof=stdin; /* input from stdin */
    }
  else{ /* input from file */
    (void)fprintf(stdout,
        "Reading profile data from profile file %s.\n",
        filename_prof);
    if((fp_prof=fopen(filename_prof,"r"))==(FILE*)NULL){
      FATAL_ERROR(E_BAD_INFILE,
          "Profile file could not be opened");
      }
    }

  /* Input the profile data */
  if(input_profile(fp_prof,&system_parm)!=0){
    FATAL_ERROR(E_BAD_DATA,
        "Error inputting system parameters and run options");
    }

  /* Close the profile file unless inputting from keyboard */
  if(fp_prof!=stdin){
    (void)fclose(fp_prof);
    fp_prof=(FILE*)NULL;
    }


  /* Open output file */
  if(filename_out[0]=='-'&&filename_out[1]=='\0'){ /* stdout */
    (void)fprintf(stdout,
        "Data will be output to stdout.\n");
    fp_out=stdout; /* output to stdout */
    }
  else{ /* output to file */
    if((fp_out=fopen(filename_out,"w"))==(FILE*)NULL){
      FATAL_ERROR(E_BAD_OUTFILE,
          "Output file could not be opened");
      }
    /* Add a line to the file indicating the date and time */
    fprintf(fp_out,"%s",ctime(&run_time));
    }


  /* Open diagnostic file to verify that it can be opened OK and
     to erase its previous contents.  This file will be opened and
     closed as data is written so the file can be viewed during
     program operation.  (The opening and closing will not be done
     if stderr is selected since these actions are not allowed for
     stderr. */
  if(filename_diag[0]=='-'&&filename_diag[1]=='\0'){ /* stderr */
    (void)fprintf(stdout,
        "Diagnostics will be sent to stderr.\n");
    fp_diag=stderr; /* input from keyboard */
    }
  else{ /* diagnostic file */
    if((fp_diag=fopen(filename_diag,"w"))==(FILE*)NULL){
      FATAL_ERROR(E_BAD_OUTFILE,
          "diagnistic file could not be opened");
      }

    /* Add the current time and date to the file and then close */
    fprintf(fp_diag,"%s",ctime(&run_time));
    (void)fclose(fp_diag);
    }


  /* Input options */
  /* In this version of the program, options consist of
       number of grid points
       match sec coil to grid, 0=no, 1=yes
       print grid, 0=no, 1=yes
  */

  /* Open option file */
  if(filename_opt[0]=='-'&&filename_opt[1]=='\0'){ /* stdin */
    (void)fprintf(stdout,
        "Options will be input from keyboard.\n");
    fp_opt=stdin; /* input from keyboard */
    }
  else{ /* data file */
    if((fp_opt=fopen(filename_opt,"r"))==(FILE*)NULL){
      FATAL_ERROR(E_BAD_INFILE,
          "Option file could not be opened");
      }
    }

  /* Input the option data */
  if(input_option(fp_opt,&option)!=0){
    FATAL_ERROR(E_BAD_DATA,
        "Error inputting options");
    }

  /* Close the option file */
  if(fp_opt!=stdin)(void)fclose(fp_opt);



  /* Input room parameters */
  /* Open room file */
  if(filename_room[0]=='-'&&filename_room[1]=='\0'){ /* stdin */
    (void)fprintf(stdout,
        "Room data will be input from keyboard.\n");
    fp_room=stdin; /* input from keyboard */
    }
  else{ /* data file */
    if((fp_room=fopen(filename_room,"r"))==(FILE*)NULL){
      FATAL_ERROR(E_BAD_INFILE,
          "Room file could not be opened");
      }
    }

  /* Input the room data */
  if(input_room(fp_room,&grid->room)!=0){
    FATAL_ERROR(E_BAD_DATA,
        "Error inputting room parameters");
    }

  /* Close the room file */
  if(fp_room!=stdin)(void)fclose(fp_room);



  /* Input coil parameters */
  /* Open coil files */
  if(filename_pri[0]=='-'&&filename_pri[1]=='\0'){ /* stdin */
    (void)fprintf(stdout,
        "Primary coil data will be input from keyboard.\n");
    fp_pri=stdin; /* input from keyboard */
    }
  else{ /* data file */
    if((fp_pri=fopen(filename_pri,"r"))==(FILE*)NULL){
      FATAL_ERROR(E_BAD_INFILE,
          "Primary coil file could not be opened");
      }
    }
  if(filename_sec[0]=='-'&&filename_sec[1]=='\0'){ /* stdin */
    (void)fprintf(stdout,
        "Secondary coil data will be input from keyboard.\n");
    fp_sec=stdin; /* input from keyboard */
    }
  else{ /* data file */
    if((fp_sec=fopen(filename_sec,"r"))==(FILE*)NULL){
      FATAL_ERROR(E_BAD_INFILE,
          "Secondary coil file could not be opened");
      }
    }

  /* Input coil data */
  if(input_coil(fp_pri,fp_sec,&grid->room,&coil)!=0){
    FATAL_ERROR(E_BAD_DATA,
        "Error inputting coil parameters");
    }

  /* Close the primary and secondary coil files */
  if(fp_pri!=stdin)(void)fclose(fp_pri);
  if(fp_sec!=stdin)(void)fclose(fp_sec);



  /* Calculate the number of points in each direction so that the
     number of grid points minimally meets or exceeds the user
     specification based on the shape of the room.

     (n_z-1)/(n_r-1) = z/r -> n_z = (n_r-1)*z/r +1
     n_r*n_z = n_grid

     n_z-1 and n_r-1 are used since there are n-1 intervals with n
     points.
     
     Substitute and solve quadratic equation.
  */
  {
    float z_r = grid->room.z/grid->room.r;
                                                /* z/r ratio */
    float b = 1-z_r; /* "b" in quadratic formula */

    grid->room.n_r = (int)ceil(
      (-b+sqrt(SQ(b)+4*z_r*(float)option.n_grid_point))/(2*z_r));

    /* Make the room radius smaller to force the secondary coil radius
       to lie exactly on a vertical grid line if this is requested */
    if(option.match_grid_to_sec!=0){
      float dl_temp = grid->room.r/(float)(grid->room.n_r-1);
      grid->room.r = grid->room.r*
          (coil.sec_r/dl_temp)/(float)ceil(coil.sec_r/dl_temp);
      }

    /* Find the number of points in the z direction.  Preserve aspect
       ratio.  Since n_r is rounded up, the number of grid points will
       be met or exceeded. */
    grid->room.n_z = (int)ceil((float)(grid->room.n_r-1)*z_r) +1;
    grid->room.dl = grid->room.r/(float)(grid->room.n_r-1);
                                   /* step= r/#pts in r dir -1 */
    }
    
  /* The rounding up above will result in a discrete room model which
     is the same width as the original but possibly taller.  Any added
     height will be put between the terminal and the ceiling. */



  /* Find descritized equivalents of the coil.  The MINs prevent
     the coil from touching the ceiling or wall */
  coil.n_pri_h_i = MIN(grid->room.n_z-2,
      (int)((coil.pri_h_i+grid->room.dl/2)/grid->room.dl));
  coil.n_pri_h_o = MIN(grid->room.n_z-2,
      (int)((coil.pri_h_o+grid->room.dl/2)/grid->room.dl));
  coil.n_pri_r_i = MIN(grid->room.n_r-2,
      (int)((coil.pri_r_i+grid->room.dl/2)/grid->room.dl));
  coil.n_pri_r_o = MIN(grid->room.n_r-2,
      (int)((coil.pri_r_o+grid->room.dl/2)/grid->room.dl));
  coil.n_sec_h_b = MIN(grid->room.n_z-2,
    (int)((coil.sec_h_b+grid->room.dl/2)/grid->room.dl));
  coil.n_sec_h_t = MIN(grid->room.n_z-2,
      (int)((coil.sec_h_t+grid->room.dl/2)/grid->room.dl));
  coil.n_sec_r = MIN(grid->room.n_r-2,
      (int)((coil.sec_r+grid->room.dl/2)/grid->room.dl));
  coil.n_term_h = MIN(grid->room.n_r-2,
      (int)((coil.term_h+grid->room.dl/2)/grid->room.dl));



  /* Allocate memory to define the first grid */
  if(allocate_grid_memory(grid)!=0){
    FATAL_ERROR(E_GRID_ALLOC,
      "Error allocating grid memory");
    }

  /* Initialize the row pointers */
  {
    int i; /* index */
    for(grid->v_row[0]=grid->v,
        grid->border_guard_row[0]=grid->border_guard,
        i=1;i<grid->room.n_z;i++){
      grid->v_row[i] =
          grid->v_row[i-1]+grid->room.n_r;
      grid->border_guard_row[i]=grid->border_guard_row[i-1]+
          grid->room.n_r;
      }
    } /* end of block intializing row pointers */



  /**** Initialize voltage and border_guard matrix ****/
  {
    int i,j;

    /* Init guard matrix to 0 */
    for(i=0;i<grid->room.n_r*grid->room.n_z;i++){
      grid->border_guard[i] = (GUARD_TYPE)0;
      }

    /* Fill in the borders.  It is not actually necessary to fill in
       the wall, floor, or ceiling, since these regions are
       automatically omitted, but this will be done anyhow since it
       adds negligible exection time and may prevent errors if part
       of the program is modified at a later date and the undefined
       border_guard points are forgotten.  Similiarly, the corners
       of the room will be assigned multiple times (only 2 extra
       stores for this). */

    /* Floor */
    for(i=0;i<grid->room.n_r;i++){
      grid->border_guard[i] = (GUARD_TYPE)1;
      grid->v[i] = 0.0f;
      }

    /* Ceiling */
    for(i=grid->room.n_r*(grid->room.n_z-1);
        i<grid->room.n_r*grid->room.n_z;i++){
      grid->border_guard[i] = (GUARD_TYPE)1;
      grid->v[i] = 0.0f;
      }

    /* Wall */
    for(i=grid->room.n_r-1;
        i<grid->room.n_r*grid->room.n_z;i+=grid->room.n_r){
      grid->border_guard[i] = (GUARD_TYPE)1;
      grid->v[i] = 0.0f;
      }

    /* Primary */
    if(coil.n_pri_h_i==coil.n_pri_h_o){ /* pancake */
      for(i=coil.n_pri_r_i, /* step horizontally */
          j=coil.n_pri_h_i*grid->room.n_r+coil.n_pri_r_i;
          i<=coil.n_pri_r_o; i++,j++){
        grid->border_guard[j]=(GUARD_TYPE)1;
        grid->v[j] = 0.0f;
        }
      }
    else if(coil.n_pri_r_i==coil.n_pri_r_o){ /* solenoid */
      for(i=coil.n_pri_h_i, /* step vertically */
          j=coil.n_pri_h_i*grid->room.n_r+coil.n_pri_r_i;
          i<=coil.n_pri_h_o; i++,j+=grid->room.n_r){
        grid->border_guard[j]=(GUARD_TYPE)1;
        grid->v[j] = 0.0f;
        }
      }
    else{ /* general cone */
      for(i=coil.n_pri_h_i; /* fill vertically */
           i<=coil.n_pri_h_o; i++){
        /* Find horizontal value */
        j = MIN(grid->room.n_r-2,
            (int)((coil.pri_r_i+
                (coil.pri_r_o-coil.pri_r_i)*
                ((float)i*grid->room.dl-coil.pri_h_i)/
                (coil.pri_h_o-coil.pri_h_i)+grid->room.dl/2)/
                grid->room.dl));
        j = i*grid->room.n_r + j;
        grid->border_guard[j]=(GUARD_TYPE)1;
        grid->v[j] = 0.0f;
        }
      for(i=coil.n_pri_r_i; /* do again horiz for better fill */
           i<=coil.n_pri_r_o; i++){
        /* Find vertical value */
        j = MIN(grid->room.n_z-2,
            (int)((coil.pri_h_i+
                (coil.pri_h_o-coil.pri_h_i)*
                ((float)i*grid->room.dl-coil.pri_r_i)/
                (coil.pri_r_o-coil.pri_r_i)+grid->room.dl/2)/
                grid->room.dl));
        j = j*grid->room.n_r + i;
        grid->border_guard[j]=(GUARD_TYPE)1;
        grid->v[j] = 0.0f;
        }
      }

    /* Secondary ground */
    if(option.draw_sec_ground!=0){
      for(i=0;i<=coil.n_sec_h_b*grid->room.n_r; i+=grid->room.n_r){
        grid->border_guard[i] = (GUARD_TYPE)1;
        grid->v[i] = 0.0f;
        }
      }

     /* Secondary coil */
    for(i=coil.n_sec_h_b*grid->room.n_r+coil.n_sec_r,
            j=coil.n_sec_h_b;
        i<=coil.n_sec_h_t*grid->room.n_r+coil.n_sec_r;
        i+=grid->room.n_r,j++){
      grid->border_guard[i] = (GUARD_TYPE)1;
      grid->v[i] = (float)sin(PI/2*(float)(j-coil.n_sec_h_b)/
            (float)(coil.n_sec_h_t-coil.n_sec_h_b));
      }

    /* Terminal */

    /* Open terminal file */
    if(filename_term[0]=='-'&&filename_term[1]=='\0'){ /* stdin */
      (void)fprintf(stdout,
          "Primary coil data will be input from keyboard.\n");
      fp_term=stdin; /* input from keyboard */
      }
    else{ /* data file */
      if((fp_term=fopen(filename_term,"r"))==(FILE*)NULL){
        FATAL_ERROR(E_BAD_INFILE,
            "Terminal data file could not be opened");
        }
      }
    
    /* input data defining the terminal */
    if(input_terminal(fp_term,grid,&coil)!=0){
      FATAL_ERROR(E_NO_MEMORY,
        "error inputting terminal data");
      }
    /* Close the terminal file */
    if(fp_term!=stdin)(void)fclose(fp_term);

    } /* end of block intializing arrays */



  /* Generate horizontal ranges for each row from border data */
  {
    int i,j;
    int n_range=0; /* total number of ranges required */
    GUARD_TYPE *border_cur; /* pointer to current border row */
    struct Range *range_cur; /* pointer to current range data */
    for(i=0;i<grid->room.n_z;i++){
      for(grid->range_h_count[i]=0,
          border_cur=grid->border_guard_row[i],
          j=grid->room.n_r-1;j>0; ){ /* r=0 pt excluded */

        /* Start at wall, a known border point and step to center
           until a non-border point is reached */
        for(j--;(j>0)&&(border_cur[j]!=(GUARD_TYPE)0);j--);

        /* Verify that the point reached is not guarded (instead
           of being end of range r=1) */
        if(j>0 && border_cur[j]==(GUARD_TYPE)0){
          grid->range_h_count[i]++; /* the point starts a new range
               (or ends one if considering from center) */

          /* Start at beginning of range and step to center until
             a border point is reached */
          for(j--;(j>0)&&(border_cur[j]==(GUARD_TYPE)0);j--);
          } /* end of new unguarded range */
        } /* end of loop scanning points in a single row */
      n_range += grid->range_h_count[i]; /* update range count */
      } /* end of loop to scan a single row */

    /* Allocate range storage based on the largest number of ranges
       found in a single row */
    if((grid->range_h=
        (struct Range*)malloc(n_range*sizeof(struct Range)))==
        (struct Range*)NULL){
      FATAL_ERROR(E_NO_MEMORY,
          "memory for horizontal range storage could not be "
          "allocated");
      }

    /* Initialize pointers to range rows */
    for(grid->range_h_row[0]=grid->range_h,
        i=1;i<grid->room.n_z;i++){
      grid->range_h_row[i]=grid->range_h_row[i-1]+
          grid->range_h_count[i-1];
      }


    /* Go through the array again, and this time actually set
       the ranges.  Again, the scan will be from wall (a known
       border point) to center. */
    for(i=0;i<grid->room.n_z;i++){
      for(range_cur=grid->range_h_row[i]+grid->range_h_count[i]-1,
                                  /* last range in this vertical row */
          border_cur=grid->border_guard_row[i],
          j=grid->room.n_r-1;j>0; ){

        /* Start at wall, a known border point and step to center
           until a non-border point is reached */
        for(j--;(j>0)&&(border_cur[j]!=(GUARD_TYPE)0);j--);

        /* Verify that the point reached is not guarded (instead
           of being just the endpoint) */
        if(j>0 && border_cur[j]==(GUARD_TYPE)0){
          range_cur->high=j; /* the point is one closest to wall in
                                the current range */

          /* Start at beginning of range and step to center until
             a border point is reached */
          for(j--;(j>0)&&(border_cur[j]==(GUARD_TYPE)0);j--);

          /* Set the start of the range.  If the point is not guarded
             then it is (an unguarded) centerline point.  Otherwise,
             it is the start of a new guarded region. 

             The low end of this interval is still set to 1 since
             the centerline is a special case that must be considered
             separately because there is no "V[-1]" term and the
             "V[1]" term is used in its place by symmetry. */
          if(j==0)range_cur->low=1;
          else range_cur->low=j+1;
          range_cur--; /* set next range closer to center */
          } /* end of new unguarded range */
        } /* end of loop scanning points in a single row */
      } /* end of loop to scan a single row */

    } /* end of block to generate ranges */


    /* Initialize non-border points of voltage matrix.
       The current method is to set these to 0 */
  {
    int i;
    for(i=0;i<grid->room.n_r*grid->room.n_z;i++){
      if(grid->border_guard[i]==(GUARD_TYPE)0)grid->v[i] = 0.0f;
      }
    } /* end of block to initialize voltage matrix */

  /* Copy first grid to duplicates.  The duplicates use the same
     memory EXCEPT for the voltage and voltage row arrays. */
  /* Unless there are >1 grids, this loop will not execute */
  {
    int i,j;
    for(i=1;i<N_GRID;i++){
      memcpy(grid+i,grid+i-1,sizeof(struct Grid_data));
      if(allocate_voltage_memory(grid+i)!=0){
        FATAL_ERROR(E_NO_MEMORY,
          "memory for duplicate voltage array could not be allocated");
        }

      /* Copy the first v matrix to duplicates */
      memcpy(grid[i].v,grid->v,
          grid->room.n_r*grid->room.n_z*sizeof(float));
      for(grid[i].v_row[0]=grid[i].v,j=1;j<grid->room.n_z;j++){
        grid[i].v_row[j] = grid[i].v_row[j-1]+grid->room.n_r;
        }
      } /* end of loop allocating and copying remaining V arrays */
    } /* end of block */

  /* Output room data to output file if requested */
  if(option.print_room!=0)print_room(fp_out,grid,option.print_room);

  /* Output coil data to output file if requested */
  if(option.print_coil!=0)print_coil(fp_out,&coil,option.print_coil);

  /* Output guard matrix data to output file if requested */
  if(option.print_guard!=0)print_guard(fp_out,grid->border_guard_row,
      grid->room.n_z,grid->room.n_r);

  /* Output range data to output file if requested */
  if(option.print_range!=0)print_range(fp_out,grid,option.print_range);



  /* Perform iterations if they are needed */
  if(option.print_volt!=0||option.print_cap!=0||option.print_freq!=0){
    int i; /* iteration counter */
    float residual_time=0.0f; /* elapsed time with residual testing */
    clock_t t1,t2; /* time points */
    int converge; /* has problem converged? */
    int n_iter_required; /* number of iterations required */

#ifdef DEBUG_freq
  if(compare_v(filename_diag,"before iterations:",grid,grid)!=0){
    FATAL_ERROR(999,
        "Error while comparing voltages");
    }
#endif


    t1 = clock();
    for(i=0;i<option.n_iter_max;i++){
      /* Do 1 iteration in iterative solution to Laplace's equation */
      laplace_iter_interval_c(grid,option.residual_max,&converge);
      if(converge!=0){ /* if problem converged, then exit */
        t2 = clock();
        residual_time=(float)(t2-t1);
        n_iter_required=i+1;
        goto done;
        }
      } /* end of loop of iterations */


    /* The problem failed to converge after the maximum allowed number
       of iterations if this point is reached */
    t2 = clock();
    residual_time=(float)(t2-t1);
    n_iter_required=option.n_iter_max+1; /* flag failure */

done:

    /* Print convergence results */
    if(n_iter_required<=option.n_iter_max){ /* converged */
      (void)fprintf(stdout,
          "The problem converged in %d iterations.\n",
          n_iter_required);
      }
    else{
      (void)fprintf(stdout,
          "The problem did not converge after %d iterations.\n",
          n_iter_required-1);
      }

    /* Print CPU usage */
    if(clock()==(clock_t)-1){ /* no timing facilities */
      (void)fprintf(stdout,
          "Timing facilites were not avaliable.\n");
      }
    else{ /* print CPU time in seconds */
      (void)fprintf(stdout,
        "With residual testing, iterations took %f seconds CPU time.\n"
        "Warning:  CPU times may be in error if the program has "
            "used more than\n"
        "     %f minutes of CPU time.\n",
      residual_time/(float)CLOCKS_PER_SEC,
      (float)pow(2,sizeof(clock_t)*8)/CLOCKS_PER_SEC/60);
           /* This assumes 8 bits per byte.  Function pow() is used
              instead of shifting because it is unknown (in general)
              how numbers are stored -- big-endian vs little-endian */
      } /* end of else print CPU time */

    /* Print voltage data if requested */
    if(option.print_volt!=0)print_volt(fp_out,grid->v_row,
        grid->room.n_z,grid->room.n_r,option.print_volt);

    } /* end block for iterations */

  /* Calculate cap value if needed */
  if(option.print_cap!=0 || option.print_freq!=0){
  /* Let the length from one grid point to its nearest neighbor be gu.

     As a first try for C, I will use a surface just inside of the
     room and ignore the primary coil since the field looks to be
     negligible around it and ignoring it simplifies the surface.  I
     will also assume full V for all Q.  The surface will more
     preciesly be a cylinder 0.5 gu from the boundary defined by the
     room.

     The surface is then a cylinder with radius room.n_r-1.5 gu and
     room.n_z-2 gu in height.  Since the wall, floor and ceiling are
     conducting, D is normal on this entire surface.

     For the wall, surface integral D . dS .= sum from i=1 to
     grid->room.n_z-2 of (D*2*PI*grid->room.n_r-1.5)*1 gu*gu.
     
     For the ceiling and floor, surface integral D . dS .= sum from
     i=1 to grid->room.n_r D*PI*(Ro2-Ri2) gu*gu.

     D will be approximated by the values equidistant from the surface
     and perpendicular from it. */

    int i;
    float c; /* cap value of sec */
    float period; /* 1/frequency */
    float flux_side;
    float flux_top_bottom;
    float r_prev_sq, r_cur_sq; /* radii squared */
    char work_string[SI_STRING_SIZE];
    char *si_string; /* ptr to number conv to number + SI prefix */

    /* wall */

    /* Find flux passing through the side of the cylinder.  A
       factor of 2*PI*R*e0 is not included yet */
    for(flux_side=0.0f,i=1;i<grid->room.n_z-1;i++){
      flux_side += grid->v_row[i][grid->room.n_r-2];
                /* V is 0 on boundary, units are V/gu*gu = V */
      }

    /* Find flux passing through the floor and ceiling of the
       cylinder.  A factor of PI*e0 is not included yet */
    for(flux_top_bottom=0.0f,r_prev_sq=0.0f,i=0;
        i<grid->room.n_r-1;i++){
      r_cur_sq=SQ((float)i+0.5f);
      flux_top_bottom +=
          (grid->v_row[grid->room.n_z-2][i] +
          grid->v_row[1][i])*
              (r_cur_sq-r_prev_sq); /* 0V on boundary of room */
      r_prev_sq=r_cur_sq; /* units are V*gu */
      }

    c = (2.0f*flux_side*((float)grid->room.n_r-1.5f)+
        flux_top_bottom)* /* units of V*gu here */
           (float)PI*grid->room.dl* /*units of V*m */
           (float)E0; /*units of V*m*F/m = V*F = F since V=1 */

    /* Output capacitance if requested */
    if(option.print_cap!=0){
      si_string=si_prefix(c,option.print_cap,work_string);
      (void)fprintf(fp_out,
          "Secondary lumped C= %sF\n",si_string);
      if(fp_out!=stdout){
        (void)fprintf(stdout,
          "Secondary lumped C= %sF\n",si_string);
        }
      }

    /* Output frequency if requested */
    if(option.print_freq!=0){
      /* Find period and ensure it is non-zero */
      period = 2.0f*(float)PI*(float)sqrt(coil.sec_l*c);
      if(period<1E-20f){
        FATAL_ERROR(E_DIVIDE_BY_0,
            "Resonant frequency is infinite??");
        }

      si_string=si_prefix(1.0/period,option.print_freq,work_string);
      (void)fprintf(fp_out,
          "Resonant frequency= %sHz\n",si_string);
      if(fp_out!=stdout){
        (void)fprintf(stdout,
          "Resonant frequency= %sHz\n",si_string);
        }
      } /* end of request to print resonant frequency */
    } /* end of block finding C */


exit:
  /* Clean up and exit */
    
  /* Close any remaining open files */
  if(fp_prof!=(FILE*)NULL){
    (void)fclose(fp_prof);
    }
  if(fp_opt!=stdin && fp_opt!=(FILE*)NULL){
    (void)fclose(fp_opt);
    }
  if(fp_room!=stdin && fp_room!=(FILE*)NULL){
    (void)fclose(fp_room);
    }
  if(fp_pri!=stdin && fp_pri!=(FILE*)NULL){
    (void)fclose(fp_pri);
    }
  if(fp_sec!=stdin && fp_sec!=(FILE*)NULL){
    (void)fclose(fp_sec);
    }
  if(fp_term!=stdin && fp_term!=(FILE*)NULL){
    (void)fclose(fp_term);
    }
  if(fp_out!=stdout && fp_out!=(FILE*)NULL){
    (void)fclose(fp_out);
    }
  if(fp_diag!=stderr && fp_diag!=(FILE*)NULL){
    (void)fclose(fp_diag);
    }


  /* Free all the memory associated with the grid(s). */
  {
    int i;
    if(grid->border_guard!=(GUARD_TYPE*)NULL)
        free(grid->border_guard);
    if(grid->border_guard_row!=(GUARD_TYPE**)NULL)
        free(grid->border_guard_row);
    if(grid->range_h!=(struct Range*)NULL)free(grid->range_h);
    if(grid->range_h_row!=(struct Range**)NULL)
        free(grid->range_h_row);
    if(grid->range_h_count!=(int*)NULL)free(grid->range_h_count);

    /* Free v arrays and row pointers from other grids, which have
       copies of these items. */
    for(i=0;i<N_GRID;i++){
      if(grid[i].v!=(float*)NULL)free(grid[i].v);
      if(grid[i].v_row!=(float**)NULL)free(grid[i].v_row);
      }

    } /* end of block freeing grid memory */


  return xrc;
  } /* end of main program */


/**********************************************************************
   This function inputs paramters for the overall system and
   paramters for the run 
**********************************************************************/
int input_profile(FILE *fp_prof, struct System_parm *system_parm){

  /* Input the option file name */
  if(input_string_echo(fp_prof,
      "enter the name of the option file, - for keyboard",
      system_parm->filename_opt)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting the name of the option file");
    }

  /* Input the room file name */
  if(input_string_echo(fp_prof,
      "enter the name of the room file, - for keyboard",
      system_parm->filename_room)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting the name of the room file");
    }

  /* Input height of primary inner turn and the primary file name */
  if(input_length_echo(fp_prof,
      "height of innermost primary turn above floor",
      system_parm->pri_h_i)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting primary inner winding");
    }


  if(input_string_echo(fp_prof,
      "enter the name of the file describing the primary coil, - for "
          "keyboard",
      system_parm->filename_pri)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting the file describing the primary coil");
    }

  /* Input height of secondary base above floor and secondary file
     name */
  if(input_length_echo(fp_prof,
      "height of bottom of secondary coil above floor",
      system_parm->sec_h_b)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting height of secondary base");
    }

  if(input_string_echo(fp_prof,
      "enter the name of the file describing the secondary, - for "
          "keyboard",
      system_parm->filename_sec)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting the file describing the secondary coil");
    }

  /* Input height of terminal base above floor and terminal file
     name */
  if(input_length_echo(fp_prof,
      "height of bottom of terminal above floor",
      system_parm->term_h)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting height of secondary base");
    }

  if(input_string_echo(fp_prof,
      "enter the name of the file describing the terminal, - for "
          "keyboard",
      system_parm->filename_term)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting the name of the file defining the terminal");
    }

  /* Input the output file name */
  if(input_string_echo(fp_prof,
      "enter the name of the output file, - for screen (stdout)",
      system_parm->filename_out)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting the name of the output file");
    }

  /* Input the diagnostic name */
  if(input_string_echo(fp_prof,
      "enter the name of the diagnostic file, - for screen (stderr)",
      system_parm->filename_diag)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting the name of the diagnostic file");
    }

  return 0;
  }



/**********************************************************************
   This function inputs options for the program.
**********************************************************************/
int input_option(FILE *fp_in,struct Option *option){

  if(input_int_echo_range(fp_in,"number of grid points",
      &option->n_grid_point,4,10000000)!=0){ /* resolution */
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting number of grid points");
    }

  if(input_int_echo_range(fp_in,"match grid to secondary? "
      "(0=no, 1=yes)",
      &option->match_grid_to_sec,0,1)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting option to match grid to secondary");
    }

  if(input_int_echo_range(fp_in,"draw a ground from secondary? "
      "(0=no, 1=yes (normally 1))",
      &option->draw_sec_ground,0,1)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting option to draw a ground from secondary");
    }

  if(input_float_echo_range(fp_in,"enter maximum allow residual",
      &option->residual_max,0.0f,10000.0f)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting maximum allowable residual");
    }

  if(input_int_echo_range(fp_in,"enter maximum number of iterations",
      &option->n_iter_max,1,100000000)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting maximum number of iterations");
    }

  if(input_int_echo_range(fp_in,"print room data? "
      "(0=no, 1<i<="PRECISION_MAX_STRING"=yes, with precision i)",
      &option->print_room,0,PRECISION_MAX)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting option to print room data");
    }

  if(input_int_echo_range(fp_in,"print coil data? "
      "(0=no, 1<i<="PRECISION_MAX_STRING"=yes, with precision i)",
      &option->print_coil,0,PRECISION_MAX)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting option to print coil data");
    }

  if(input_int_echo_range(fp_in,"print guard matrix data? "
      "(0=no, 1=yes)",
      &option->print_guard,0,1)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting option to print guard matrix data");
    }

  if(input_int_echo_range(fp_in,"print range data? "
      "(0=no, 1=yes)",
      &option->print_range,0,1)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting option to print range data");
    }

  if(input_int_echo_range(fp_in,"print final voltage data? "
      "(0=no, 1<i<="PRECISION_MAX_STRING"=yes, scaled by 10Ei)",
      &option->print_volt,0,PRECISION_MAX)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting option to output final voltage data");
    }

  if(input_int_echo_range(fp_in,"print capacitance of coil? "
      "(0=no, 1<i<="PRECISION_MAX_STRING"=yes, with precision i)",
      &option->print_cap,0,PRECISION_MAX)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting option to output capacitance of coil");
    }

  if(input_int_echo_range(fp_in,"print resonant frequency? "
      "(0=no, 1<i<="PRECISION_MAX_STRING"=yes, with precision i)",
      &option->print_freq,0,PRECISION_MAX)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting option to output resonant frequency");
    }

  return 0;
  } /* end of function input_option */



/**********************************************************************
   This function inputs room dimensions.
**********************************************************************/
int input_room(FILE *fp_in,struct Room *room){

  if(input_length_echo_range(fp_in,"distance from coil to wall",
      &room->r,0.001f,10000.0f)!=0){ /* coil to wall */
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting distance from coil to wall");
    }
  if(input_length_echo_range(fp_in,"height of ceiling",&room->z,
      0.001f,10000.0f)!=0){ /* ceiling height */
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting ceiling height");
    }

  return 0;
  } /* end of function input_room */



/**********************************************************************
   This function allocates memory for voltage and its row matrix in a
   grid.
**********************************************************************/
int allocate_voltage_memory(struct Grid_data *grid){
  if((grid->v=(float*)
      malloc(grid->room.n_r*grid->room.n_z*sizeof(float)))==
      (float*)NULL){
    FUNCTION_ERROR(E_NO_MEMORY,
        "Memory for voltage array could not be allocated");
    }
  if((grid->v_row=
      (float**)malloc(grid->room.n_z*sizeof(float*)))==
      (float**)NULL){
    FUNCTION_ERROR(E_NO_MEMORY,
        "Memory for voltage array location storage could not be "
        "allocated");
    }
  return 0;
  } /* end of function allocate_voltage_memory */



/**********************************************************************
   This function allocates memory for various components defining a
   grid.
**********************************************************************/
int allocate_grid_memory(struct Grid_data *grid){

  if(allocate_voltage_memory(grid)!=0){
    FUNCTION_ERROR(E_NO_MEMORY,
        "Memory for voltage array or its row pointers could not be "
        "allocated");
    }

  if((grid->border_guard=
      (GUARD_TYPE*)
        malloc(grid->room.n_r*grid->room.n_z*sizeof(GUARD_TYPE)))==
      (GUARD_TYPE*)NULL){
    FUNCTION_ERROR(E_NO_MEMORY,
        "memory for guard storage could not be allocated");
    }

  if((grid->border_guard_row=
      (GUARD_TYPE**)
        malloc(grid->room.n_r*grid->room.n_z*sizeof(GUARD_TYPE*)))==
      (GUARD_TYPE**)NULL){
    FUNCTION_ERROR(E_NO_MEMORY,
        "memory for guard-row storage could not be allocated");
    }

  if((grid->range_h_count=
      (int*)malloc(grid->room.n_z*sizeof(int)))==(int*)NULL){
    FUNCTION_ERROR(E_NO_MEMORY,
        "memory for horizontal range count storage could not be "
        "allocated");
    }

  if((grid->range_h_row=
      (struct Range**)malloc(grid->room.n_z*sizeof(struct Range*)))==
      (struct Range**)NULL){
    FUNCTION_ERROR(E_NO_MEMORY,
        "memory for horizontal range row storage could not be "
        "allocated");
    }

  return 0;
  } /* end of function allocate_grid_memory */



/**********************************************************************
   This function inputs coil data.
**********************************************************************/
  int input_coil(FILE *fp_pri, FILE *fp_sec, struct Room *room,
      struct Coil *coil){

  /* NOTE:  The MINs are to prevent possible out-of-bounds conditions
     due to rounding during computations. */

  /* Verify height of primary winding above floor is OK */
  if(check_range_float(coil->pri_h_i,0.0f,room->z)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Invalid height for primary inner winding");
    }

  /* Height of primary outer winding above floor.  Input as height
     above inner turn and then convert.  The range testing forces
     the primary to be a pancake or spiral UPWARDS.  This is done
     to make subsequent testing for overlap of primary and secondary
     easier.  The excluded cases are very uncommon. */
  if(input_length_echo_range(fp_pri,
      "height of outermost primary turn above innermost turn",
      &coil->pri_h_o,0.0f,room->z-coil->pri_h_i)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting height of primary outer winding");
    }
  coil->pri_h_o += coil->pri_h_i; /* now height above floor */

  /* Inner primary radius.  Input as diameter and then convert */
  if(input_length_echo_range(fp_pri,
    "inner diameter of primary (parallel to floor)",
      &coil->pri_r_i,0.0f,2*room->r)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting inner radius of primary");
    }
  coil->pri_r_i = coil->pri_r_i/2; /* diameter -> radius */

  /* Outer primary radius.  Input as diameter and then convert */
  if(input_length_echo_range(fp_pri,
    "outer diameter of primary (parallel to floor)",
      &coil->pri_r_o,2*coil->pri_r_i,2*room->r)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting outer radius of primary");
    }
  coil->pri_r_o = coil->pri_r_o/2; /* diameter -> radius */

  /* Verify height of secondary base above floor is OK */
  if(check_range_float(coil->sec_h_b,0.0f,room->z)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Invalid height for secondary base");
    }

  /* Height of secondary top above floor.  Input as length of
     secondary and then convert by adding base height */
  if(input_length_echo_range(fp_sec,
      "length of secondary coil winding",
      &coil->sec_h_t,0.0f,room->z-coil->sec_h_b)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting length of secondary coil winding");
    }
  coil->sec_h_t = coil->sec_h_t + coil->sec_h_b; /* len -> ht */

  /* Input secondary radius.  Based on the height of the secondary 
     relative to the height of the primary, the range of allowable
     diameters varies */
  if(coil->sec_h_b<=coil->pri_h_i){ /* sec in pri */
    if(input_length_echo_range(fp_sec,
        "diameter of secondary (parallel to floor)",
        &coil->sec_r,0.0f,2*coil->pri_r_i)!=0){
      FUNCTION_ERROR(E_BAD_DATA,
          "Error inputting diameter of secondary");
      }
    }
  else if(coil->sec_h_b>coil->pri_h_o){ /* sec above pri */
    if(input_length_echo_range(fp_sec,
        "diameter of secondary (parallel to floor)",
        &coil->sec_r,0.0f,2*room->r)!=0){
      FUNCTION_ERROR(E_BAD_DATA,
          "Error inputting diameter of secondary");
      }
    }
  else{ /* sec in pri (conical primary) */
    if(input_length_echo_range(fp_sec,
        "diameter of secondary (parallel to floor)",
        &coil->sec_r,0.0f,
        2*(coil->pri_r_i+
            (coil->pri_r_o-coil->pri_r_i)*
            (coil->sec_h_b-coil->pri_h_i)/
            (coil->pri_h_o-coil->pri_h_i)))
            !=0){
      FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting diameter of secondary");
      }
    }
  coil->sec_r = coil->sec_r/2; /* diameter -> radius */

  /* Secondary inductance */
  if(input_float_echo_range(fp_sec,
    "Secondary inductance in Henrys",
      &coil->sec_l,0.0f,20.0f)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting secondary inductance");
    }

  return 0;
  } /* end of function input_coil */



/**********************************************************************
   This function inputs terminal data and updates the range data based
   on the data.
**********************************************************************/
int input_terminal(FILE *fp_term, struct Grid_data *grid,
    struct Coil *coil){

  float r_cur=0.0f, z_cur=coil->term_h; /* cur loc on term */
  int n_r_cur, n_z_cur;
  float r_dest, z_dest; /* destination loc */
  int n_z_dest; /* descritized destination */
  int n_r_dest; /* descritized destination */
  int n_r_new; /* new point moved to while going to destination */
  char instring[INPUT_STRING_MAX+1]; /* string to read next
                                         word from file */
  char instring_upper[INPUT_STRING_MAX+1]; /* uppercase of word */
  int i,j,k; /* index variable */
  int terminal_defined=0; /* was a terminal defined at all
                                           (vs bare coil)? */
  while(1){ /* do until exit within loop */

    /* Input next terminal shape command string */
    if(input_string_echo(fp_term,
        "terminal data (ARC/DONE/HELP/LINE/LINEH/LINEV/Q)",
        instring)!=0){
      FUNCTION_ERROR(E_BAD_DATA,
          "Error inputting terminal shape data");
      }
    uppercpy(instring_upper,instring); /* uppercase copy of input */

    
    /**** Check for done request ****/
    if(strcmp(instring_upper,"DONE")==0){

      /* Check if there were any terminal-defining commands.  If
         not, this is a bare coil and should NOT have the centerline
         from the center point of the secondary to the bottom of the
         terminal set to 1V and guarded.  Otherwise these points are
         set to 1V and guarded. */
      if(terminal_defined!=0){
        for(i=(int)((coil->sec_h_t+grid->room.dl/2)/grid->room.dl)*
                grid->room.n_r; /* locate top of secondary at center */
            i<=(int)((coil->term_h+grid->room.dl/2)/grid->room.dl)*
                grid->room.n_r; /* limit is bottom of terminal */
            i+=grid->room.n_r){ /* step along centerline */
          grid->v[i] = 1.0f;
          grid->border_guard[i]=(GUARD_TYPE)1;
          } /* end of loop init centerline between coil and term */
        }

      (void)fprintf(stdout,"Terminal input completed.\n");
      return 0; /* normal exit for the function */
      }



    /**** Check for arc request ****/
    else if(strcmp(instring_upper,"ARC")==0){
      float theta1,theta2; /* departure and arrival angles */

      /* Get final height */
      if(input_length_echo_range(fp_term,"  final height",&z_dest,
          z_cur-coil->term_h,grid->room.z-coil->term_h)!=0){
        FUNCTION_ERROR(E_BAD_DATA,
            "Error inputting final height for arc\n");
        }
      z_dest += coil->term_h; /* make relative to floor */
      n_z_dest = MIN(grid->room.n_z-1,
          (int)((z_dest+grid->room.dl/2)/grid->room.dl));

      /* Get angle of departure */
      if(input_float_echo_range(fp_term,
          "  angle of arc leaving current point",&theta1,
          0.0f,180.0f)!=0){
        FUNCTION_ERROR(E_BAD_DATA,
            "Error inputting departure angle for arc");
        }

      /* Get angle of arrival */
      if(input_float_echo_range(fp_term,
          "  angle of arc arriving at destination point",&theta2,
          0.0f,180.0f)!=0){
        FUNCTION_ERROR(E_BAD_DATA,
            "Error inputting arrival angle for arc");
        }

      /* additional range checking: destination radius cannot equal
         arrival radius.  This would result in an infinte radius */
      if(fabs(theta2-theta1)<MIN_ANGLE_DIF){
        char error_message[100];
        sprintf(error_message,
        "Departure and arrival angles must differ by more "
        "than %f degrees",MIN_ANGLE_DIF);
        FUNCTION_ERROR(E_BAD_DATA,error_message);
        }

      /* Convert degrees to radians for trig functions */
      theta1 *= (float)(PI/180);
      theta2 *= (float)(PI/180);

      /* Update range info.  An arc extends the terminal height */

      /* First the center and radius of the circle containing the
         input arc must be found from the geometry of the problem.


    (Cr,Cz)
  --------X-----------------------------------------*--- center height
     A    |\  \                                     *
     |    | \      \                                *
     |    |  \         \    radius                 *
     h    |   \            \                       *
     |    |    \               \                  *
     |    |     \                  \              *
     |    |      \                     \      90 *
     V    |       \              90-theta2 \     * theta2
  --------+------------------------------------\*-------- final height
     A    |        \                           *
     |    |         \                         *
     |    |          \ radius                *
     |    |           \                     *
     |    |<--------------- lq ----------->* 
     d    |             \                 * (Qr,Qz)
     |    |              \               *
     |    |               \             *
     |    |                \          *
     |    |                 \  90   *
     V    |        90-theta1 \    *  theta1
  --------+-------------------\*---------------------- starting height
          |<-------- l ------->|(Pr,Pz)

  geometry ->         h     = radius * cos(theta2)                   (1)
  geometry ->         h + d = radius * cos(theta1)                   (2)
  (1)/(2) ->        h/(d+h) = cos(theta2)/cos(theta1)                (3)
  (3)*(d+h)*cos(theta1) ->
              h*cos(theta1) = (h+d)*cos(theta2)                      (4)
  solving (4) for h -     h = d*cos(theta2)/(cos(theta1)-cos(theta2) (5)
  geometry ->          radius = (h+d)/cos(theta1)                    (6)
  geometry ->          radius = h/cos(theta2)                        (7)
  For improved accurracy, chose between (6) and (7) depending on which
  is furthest from 90 degrees.
  geometry ->               l =  radius*sin(theta1), theta1<theta2   (8)
  geometry ->               l =  radius*sin(theta1), theta1>theta2   (9)
                                      since radius<0 in (9)
                           Cr = Pr - l
                           Cz = Pz + h
                           Qz is specified based on grid height     
  Pythagorean theoream ->  lq = sqrt(radius*radius-(Cz-Qz)*(Cz-Qz))
                           lq = -lq if theta1>teta2
                           Qr = Cr + lq

**********************************************************************/
      {
      float d = z_dest-z_cur;
      float h = d*(float)(cos(theta2)/(cos(theta1)-cos(theta2)));
      float radius = (float)((fabs(PI/2-theta1)>fabs(PI/2-theta2)) ?
                        (h+d)/cos(theta1) : h /cos(theta2));
      float l = (float)(radius*sin(theta1));
      float r_center = r_cur - l;
      float z_center = z_cur + h + d;
      float lq;

      /* Assign points in the voltage and guard arrays */
      n_z_cur = (int)((z_cur+grid->room.dl/2)/grid->room.dl);
      for(i=n_z_cur+1;i<=n_z_dest; i++){
        lq = (float)sqrt(MAX(0.0,
            (double)(SQ(radius)-
                SQ(z_center-(float)i*grid->room.dl))));
              /* The MAX is especially necessary because i may
                 actually be above the circle due to rounding while
                 descritizing the problem.  This is in addition to
                 numerical rounding considerations. */
        if(theta1>theta2) lq = -lq;
        n_r_new = (int)((r_center+lq+grid->room.dl/2)/grid->room.dl); 
                              /* convert to a discrete grid point */
        /* See if this arc point fits in the room (and does not
           touch the wall */
        if(n_r_new>grid->room.n_r-2){
          char error_message[100];
          sprintf(error_message,
              "The arc defined does not fit within the room at height "
              "%f inches from the ground\n",
              (float)i*grid->room.dl);
          FUNCTION_ERROR(E_BAD_DATA,error_message);
          }

        /* See if this arc point goes past centerline */
        if(n_r_new<0){
          char error_message[100];
          sprintf(error_message,
              "The arc defined goes past the centerline at height "
              "%f inches from the ground\n",
              (float)i*grid->room.dl);
          FUNCTION_ERROR(E_BAD_DATA,error_message);
          }

        /* Initialize the terminal and its inside to 1V and guard
           these points */
        for(j=0,k=i*grid->room.n_r;j<=n_r_new;j++,k++){
          grid->v[k] = 1.0f;
          grid->border_guard[k]=(GUARD_TYPE)1;
          }
        } /* end of loop defining arc */

      /* Update final point by finding r coordinate of final height */
      lq = (float)(sqrt(SQ(radius)-SQ(z_center-z_dest)));
      if(theta1>theta2) lq = -lq;
      r_cur = r_center+lq;
      z_cur = z_dest;
      } /* end of block updating range info */
      terminal_defined=1; /* must be some terminal defined now */
      } /* end of case of arc */



    /**** Check for horizontal line request ****/
    else if(strcmp(instring_upper,"LINEH")==0){

      /* Get final radius */
      if(input_length_echo_range(fp_term,"  final radius",&r_dest,
          0,grid->room.r)!=0){
        FUNCTION_ERROR(E_BAD_DATA,
            "Error inputting final radius for horizontal line");
        }

      n_r_dest = MIN(grid->room.n_r-2,
          (int)((r_dest+grid->room.dl/2)/grid->room.dl));


      /* Update range info.  A horizontal line does not change the
         terminal height, but it will increase the terminal extent
         if the destination is greater that the current r location */
      n_r_cur = (int)((r_cur+grid->room.dl/2)/grid->room.dl);
      n_z_cur = (int)((z_cur+grid->room.dl/2)/grid->room.dl);
      for(i=n_r_cur+1,j=n_z_cur*grid->room.n_r+i;
          i<=n_r_dest;i++,j++){
        grid->v[j] = 1.0f;
        grid->border_guard[j]=(GUARD_TYPE)1;
        }
      r_cur = r_dest;
      terminal_defined=1; /* must be some terminal defined now */
      } /* end of case of horizontal line */



    /**** Check for vertical line request ****/
    else if(strcmp(instring_upper,"LINEV")==0){

      /* Get final height */
      if(input_length_echo_range(fp_term,"  final height",&z_dest,
          z_cur-coil->term_h,grid->room.z-coil->term_h)!=0){
        FUNCTION_ERROR(E_BAD_DATA,
            "Error inputting final height for vertical line");
        }
      z_dest += coil->term_h; /* make relative to floor */

      /* Fill in voltage and guard arrays within vertical section */
      n_z_cur = (int)((z_cur+grid->room.dl/2)/grid->room.dl);
               /* min not needed since this is lower end of range */
      n_z_dest = MIN(grid->room.n_z-2,
          (int)((z_dest+grid->room.dl/2)/grid->room.dl));
      n_r_cur = MIN(grid->room.n_r-2,
          (int)((r_cur+grid->room.dl/2)/grid->room.dl));
      for(i=n_z_cur+1;i<=n_z_dest;i++){
        for(j=0,k=i*grid->room.n_r;j<=n_r_cur;j++,k++){
          grid->v[k] = 1.0f;
          grid->border_guard[k]=(GUARD_TYPE)1;
          }
        }

      /* Update current point */
      z_cur = z_dest;
      terminal_defined=1; /* must be some terminal defined now */
      } /* end of case of vertical line */



    /**** Check for general line request ****/
    else if(strcmp(instring_upper,"LINE")==0){

      /* Get final height */
      if(input_length_echo_range(fp_term,"  final height",&z_dest,
          z_cur-coil->term_h,grid->room.z-coil->term_h)!=0){
        FUNCTION_ERROR(E_BAD_DATA,
            "Error inputting final height for line");
        }
      z_dest += coil->term_h; /* make relative to floor */
      n_z_dest = MIN(grid->room.n_z-1,
          (int)((z_dest+grid->room.dl/2)/grid->room.dl));

      /* Get final radius */
      if(input_length_echo_range(fp_term,"  final radius",&r_dest,
          0,grid->room.r)!=0){
        FUNCTION_ERROR(E_BAD_DATA,
            "Error inputting final radial distance for line");
        }

      /* Update range info.  A general line extends the terminal
         height */
      n_z_cur = (int)((z_cur+grid->room.dl/2)/grid->room.dl);
      for(i=n_z_cur+1; i<=n_z_dest; i++){
        n_r_new  = MIN(grid->room.n_r-2,
            (int)((r_cur+(r_dest-r_cur)*
                ((float)i*grid->room.dl-z_cur)/
                    (z_dest-z_cur)
                    +grid->room.dl/2)/grid->room.dl));
        for(j=0,k=i*grid->room.n_r;j<=n_r_new;j++,k++){
          grid->v[k] = 1.0f;
          grid->border_guard[k]=(GUARD_TYPE)1;
          }
        }

      r_cur = r_dest;
      z_cur = z_dest;
      terminal_defined=1; /* must be some terminal defined now */
      } /* end of case of general line */



    /**** Check for quit request -- no need since input_string()
          always checks for a quit ****/



    /**** Check for help request ****/
    else if(strcmp(instring_upper,"HELP")==0){
      (void)fprintf(stdout,
          "Terminal input commands:\n"
          "ARC final_ht initial_degrees final_degrees -- creates arc\n"
          "   ARC x 0 180 makes semicircle\n"
          "DONE -- no more terminal input\n"
          "HELP -- this help message\n"
          "LINE r h -- line from cur pt to radius r, height h\n"
          "LINEH r -- horizontal line from cur pt to radius r\n"
          "LINEV r -- vertical line from cur pt to height h\n"
          "Q -- quit program\n"
          "\n"
          "All measumements are relative to bottom of the terminal\n\n"
          );
        }
    /**** Nothing left -- bad input ****/
    else{
      FUNCTION_ERROR(E_BAD_DATA,
          "Invalid terminal definiton command was input");
      }
    } /* end of loop inputting terminal data */
  } /* end of function input_terminal */



/**********************************************************************
   Function uppercpy is identical to strcpy except that it uppercases
   any lower case letters in src while performing the copy.
**********************************************************************/
char *uppercpy(
    char *dst, /* destination */
    char *src /* pointer to source string */
    ){
  char *dst_return=dst;
  while(*src!='\0'){
    *dst++=toupper(*src++); /* uppercase and copy next char */
    }
  *dst='\0'; /* add terminator */
  return dst_return;
  } /* end of function uppercpy */



/**********************************************************************
   This function inputs string data.  All actual reading of coil data
   from the data file or stdin is done here.
**********************************************************************/
int input_string(FILE *fp_in, char *instring){

  static char format_string[]="%123456789s";
  static char first_call=1;

  /* On first call initialize the format string */
  if(first_call!=0){
    first_call=0;
    sprintf(format_string,"%%%ds",INPUT_STRING_MAX);
    }

  /* Read strings until a non-comment string is found */
  while(1){
    /* Read one string */
    if(fscanf(fp_in,format_string,instring)<=0){
      FUNCTION_ERROR(E_BAD_DATA,
          "Error inputting string from file");
      }

    /* Read to end of current line if comment */
    if(*instring=='#'){ /* found start of a comment */
      if(fscanf(fp_in,"%*[^\n]\n")<0){
        FUNCTION_ERROR(E_BAD_DATA,
            "Error while skipping a comment in file");
        }
      }
    else break; /* non-comment string was found */
    }


  
  /* Check for quit request */
  if((instring[0]=='Q'||instring[0]=='q')&&instring[1]=='\0'){
    FUNCTION_ERROR(E_BAD_DATA,
        "Program is cancelled by user request");
    }

  return 0;
  } /* end of function input_string */



/**********************************************************************
   This function inputs string data and echos it.
**********************************************************************/
int input_string_echo(FILE *fp_in, char *prompt, char *instring){

  /* Print prompt message */
  (void)fprintf(stdout,"%s: ",prompt);

  if(input_string(fp_in,instring)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting string from file");
    }

  /* Print value if not from terminal.  (Terminal shows value.) */
  if(fp_in!=stdin){
    (void)fprintf(stdout,"%s\n",instring);
    }

  return 0;
  } /* end of function input_string_echo */



/**********************************************************************
   This function checks length data for allowable range.
**********************************************************************/
int check_range_length(float x, float low, float high){

  /* Check for allowable range */
  if(x<low){
    char message[100];
    (void)sprintf(message,
        "Value is too low.  Minimum value is %f (in meters).",low);
    FUNCTION_ERROR(E_BAD_DATA, message);
    }

  if(x>high){
    char message[100];
    (void)sprintf(message,
        "Value is too high.  Maximum value is %f (in meters).",high);
    FUNCTION_ERROR(E_BAD_DATA, message);
    }

  return 0;
  } /* end of function check_range_length */




/**********************************************************************
   This function inputs length data and echos it.
**********************************************************************/
int input_length_echo(FILE *fp_in, char *prompt, float *x){

  char instring[INPUT_STRING_MAX+1]; /* string to read next word */
  int rc; /* return code from function call */
  char length_unit_string[INPUT_STRING_MAX+1];
                                  /* string for units used */
  float scale_factor; /* conversion from input units to meters */

  if(input_string_echo(fp_in,prompt,instring)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error reading string from file or echoing to terminal");
    }

  /* Convert the string to a number, if possible.  Exit if not */
  rc = sscanf(instring,"%f%s",x,length_unit_string);
  if(rc<=0){ /* complete failure */
    char message[300];
    (void)sprintf(message,
        "Error converting input data %200s to a length",instring);
    FUNCTION_ERROR(E_BAD_DATA, message);
    }

  if(rc==1){ /* units not given */
    FUNCTION_ERROR(E_BAD_DATA,"Error reading length unit");
    }

  /* Find the required units and convert to meters. */
  if(length_unit_factor(length_unit_string,&scale_factor)!=0){
    char message[300];
    (void)sprintf(message,
        "Error converting input data %200s to a length",instring);
    FUNCTION_ERROR(E_BAD_DATA, message);
    }
  *x *= scale_factor; /* scale length */

  return 0;
  } /* end of function input_length_echo */



/**********************************************************************
   This function finds the conversion factor which will convert from
   the string's unit to meters.
**********************************************************************/
int length_unit_factor(const char *length_unit_string,float *factor){

  /* Structure to map strings to floating point numbers */
  struct Unit2factor{
    char unit[20];
    float factor;
    };

  /* Array of conversions from length unit to meter.  This array must
     be in alphabetical order since it will be searched with a binary
     search */
#define N_LENGTH_UNIT_DEFINED 5
  const static struct Unit2factor length_unit2length_factor[]={
      {"cm",(float)0.01},
      {"ft",(float)M_PER_FOOT},
      {"in",(float)M_PER_INCH}, 
      {"m",(float)1.0},
      {"mm",(float)0.001},
    };
  struct Unit2factor *unit_factor;

  /* Do a binary search for the length unit */
  if((unit_factor=(struct Unit2factor*)bsearch(length_unit_string,
      length_unit2length_factor,
      N_LENGTH_UNIT_DEFINED,sizeof(struct Unit2factor),strcmp))==
      (struct Unit2factor*)NULL){
    char message[300];
    (void)sprintf(message,
        "Length unit %200s is not defined",length_unit_string);
    FUNCTION_ERROR(E_BAD_DATA, message);
    }
  *factor=unit_factor->factor; /* set conversion factor */
  return 0;
  }



/**********************************************************************
   This function inputs length data, echos it, and verifies that the
   range of input is valid.
**********************************************************************/
int input_length_echo_range(FILE *fp_in, char *prompt, float *x,
    float low, float high){

  if(input_length_echo(fp_in,prompt,x)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting length data or echoing to terminal");
    }

  if(check_range_length(*x,low,high)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Length data outside of allowable range");
    }

  return 0;
  } /* end of function input_length_echo_range */



/**********************************************************************
   This function inputs float data and echos it.
**********************************************************************/
int input_float_echo(FILE *fp_in, char *prompt, float *x){

  char instring[INPUT_STRING_MAX+1]; /* string to read next word */

  if(input_string_echo(fp_in,prompt,instring)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error reading string from file or echoing to terminal");
    }

  /* Convert the string to a number, if possible.  Exit if not */
  if(sscanf(instring,"%f",x)<=0){
    char message[300];
    (void)sprintf(message,
        "Error converting input data %s to floating point",instring);
    FUNCTION_ERROR(E_BAD_DATA, message);
    }

  return 0;
  } /* end of function input_float_echo */



/**********************************************************************
   This function checks float data for allowable range.
**********************************************************************/
int check_range_float(float x, float low, float high){

  /* Check for allowable range */
  if(x<low){
    char message[100];
    (void)sprintf(message,
        "Value is too low.  Minimum value is %f.",low);
    FUNCTION_ERROR(E_BAD_DATA, message);
    }

  if(x>high){
    char message[100];
    (void)sprintf(message,
        "Value is too high.  Maximum value is %f.",high);
    FUNCTION_ERROR(E_BAD_DATA, message);
    }

  return 0;
  } /* end of function check_range_float */




/**********************************************************************
   This function inputs float data, echos it, and verifies that the
   range of input is valid.
**********************************************************************/
int input_float_echo_range(FILE *fp_in, char *prompt, float *x,
    float low, float high){

  if(input_float_echo(fp_in,prompt,x)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting floating point data or echoing to terminal");
    }

  if(check_range_float(*x,low,high)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error in range of floating point number");
    }

  return 0;
  } /* end of function input_float_echo_range */



/**********************************************************************
   This function inputs integer data and echos it.
**********************************************************************/
int input_int_echo(FILE *fp_in, char *prompt, int *x){

  char instring[INPUT_STRING_MAX+1]; /* string to read next word */

  if(input_string_echo(fp_in,prompt,instring)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error reading string from file or echoing to terminal");
    }

  /* Convert the string to a number, if possible.  Exit if not */
  if(sscanf(instring,"%d",x)<=0){
    char message[300];
    (void)sprintf(message,
        "Error converting input data %s to integer",instring);
    FUNCTION_ERROR(E_BAD_DATA, message);
    }

  return 0;
  } /* end of function input_int_echo */



/**********************************************************************
   This function checks integer data for allowable range.
**********************************************************************/
int check_range_int(int x, int low, int high){

  /* Check for allowable range */
  if(x<low){
    char message[100];
    (void)sprintf(message,
        "Value is too low.  Minimum value is %d.",low);
    FUNCTION_ERROR(E_BAD_DATA, message);
    }

  if(x>high){
    char message[100];
    (void)sprintf(message,
        "Value is too high.  Maximum value is %d.",high);
    FUNCTION_ERROR(E_BAD_DATA, message);
    }

  return 0;
  } /* end of function check_range_int */



/**********************************************************************
   This function inputs integer data, echos it, and verifies that the
   range of input is valid.
**********************************************************************/
int input_int_echo_range(FILE *fp_in, char *prompt, int *x,
    int low, int high){

  if(input_int_echo(fp_in,prompt,x)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error inputting integer data or echoing to terminal");
    }

  if(check_range_int(*x,low,high)!=0){
    FUNCTION_ERROR(E_BAD_DATA,
        "Error in range of integer");
    }

  return  0;
  } /* end of function input_int_echo_range */



/**********************************************************************
   This function performs one step in iterative solution of Laplace's
   equation using the interval method.  Convergence is tested via a
   maximum allowable residual value.
**********************************************************************/
void laplace_iter_interval_c(struct Grid_data *grid,
    float residual_max, int* converge){

  int i,j,k; /* loop indices */
  float *v_prev, *v_cur, *v_next; /* row pointers */
  struct Range *range_cur; /* pointer to current range */
  GUARD_TYPE *guard; /* pointer to guard row */
  float new_value; /* value of updated point */
  float p; /* current radius */

  /* Solve vertical row by row checking for convergence */
  for(i=1;i<grid->room.n_z-1;i++){ /* each vertical row
                                          except floor and ceiling */
    /* locate needed rows */
    v_prev = grid->v_row[i-1]; /* previous row */
    v_cur = grid->v_row[i]; /* current row */
    v_next = grid->v_row[i+1]; /* next row */
    range_cur = grid->range_h_row[i]; /* range of cur row */
    guard = grid->border_guard_row[i]; /* guard for current row */

    /* Process centerline separately since must use symmetry to get
      "-1 term.  "v[1] = v[-1]".  This could be done with another
      range along the centerline (vertically) but is done with a
      guard matrix here to facilitate performing the iterations in
      the same order as for the border method so the results can be
      compared. */
    if(guard[0]==(GUARD_TYPE)0){ /* if not a protected location */
      new_value = (4*v_cur[1]+v_prev[0]+v_next[0])/6;
      if(fabs((double)(v_cur[0]-new_value))>residual_max){
        v_cur[0] = new_value; /* assign new value */
        goto iter_only_past_centerline;
                       /* continue without convergence testing */
        }
      v_cur[0] = new_value;
      }
    
    for(j=0;j<grid->range_h_count[i];j++){ /* each range in row */
      for(k=range_cur[j].low,
              p=(float)range_cur[j].low;
          k<=range_cur[j].high;k++,p+=1.0f){
                                             /* each region in row */
          new_value = (v_cur[k-1]+v_cur[k+1]+v_prev[k]+v_next[k]+
              (v_cur[k+1]-v_cur[k-1])/(2*p))/4;
                                          /* actual iterative step */
          if(fabs((double)(v_cur[k]-new_value))>residual_max){
            v_cur[k] = new_value; /* assign new value */
            goto iter_only_in_grid;
                        /* continue without convergence testing */
            }
          v_cur[k] = new_value;
        } /* end of points in a horizontal region */
      } /* end of horizontal regions in a row */
    } /* end of vertical rows */
  *converge = 1; /* if this point is reached, all points were below
                    the required residual tolerance */
  return; /* all points have been updated */


  /* Solve vertical row by row checking for convergence */
  for(   ;i<grid->room.n_z-1;i++){ /* each vertical row except floor
                                      and ceiling */
    /* locate needed rows */
    v_prev = grid->v_row[i-1]; /* previous row */
    v_cur = grid->v_row[i]; /* current row */
    v_next = grid->v_row[i+1]; /* next row */
    range_cur = grid->range_h_row[i]; /* range of cur row */
    guard = grid->border_guard_row[i]; /* guard for current row */

    /* Process centerline separately since must use symmetry to get
      "-1 term.  "v[1] = v[-1]".  This could be done with another
      range along the centerline (vertically) but is done with a
      guard matrix here to facilitate performing the iterations in
      the same order as for the border method so the results can be
      compared. */
    if(guard[0]==(GUARD_TYPE)0){ /* if not a protected location */
      v_cur[0] = (4*v_cur[1]+v_prev[0]+v_next[0])/6;
      }

iter_only_past_centerline:    
    for(j=0;j<grid->range_h_count[i];j++){ /* each range in row */
      for(k=range_cur[j].low,
              p=(float)range_cur[j].low;
          k<=range_cur[j].high;k++,p+=1.0f){
                                             /* each region in row */
          v_cur[k] = (v_cur[k-1]+v_cur[k+1]+v_prev[k]+v_next[k]+
              (v_cur[k+1]-v_cur[k-1])/(2*p))/4;
                                          /* actual iterative step */
iter_only_in_grid:;
        } /* end of points in a horizontal region */
      } /* end of horizontal regions in a row */
    } /* end of vertical rows */
  *converge = 0; /* if this point is reached, all points were below
                    the required residual tolerance */
  return; /* all points have been updated */
  } /* end of function laplace_iter_interval_c */



/**********************************************************************
   This function prints coil data
**********************************************************************/
void print_coil(FILE *fp, struct Coil *coil, int precision){

  char work_string[7][SI_STRING_SIZE];

  (void)fprintf(fp,
      "\n\ncoil data\n"
      "  continuous versions:\n"
      "  primary height -- inner, outer ... %sm, %sm\n"
      "  primary radius -- inner, outer ... %sm, %sm\n"
      "  secondary height -- bottom, top ... %sm, %sm\n"
      "  secondary radius .................. %sm\n"
      "  discretized versions:\n"
      "  primary height -- inner, outer ... %d, %d\n"
      "  primary radius -- inner, outer ... %d, %d\n"
      "  secondary height -- bottom, top ... %d, %d\n"
      "  secondary radius .................. %d\n",
      si_prefix(coil->pri_h_i,precision,work_string[0]),
      si_prefix(coil->pri_h_o,precision,work_string[1]),
      si_prefix(coil->pri_r_i,precision,work_string[2]),
      si_prefix(coil->pri_r_o,precision,work_string[3]),
      si_prefix(coil->sec_h_b,precision,work_string[4]),
      si_prefix(coil->sec_h_t,precision,work_string[5]),
      si_prefix(coil->sec_r,precision,work_string[6]),
      coil->n_pri_h_i, coil->n_pri_h_o, coil->n_pri_r_i,
      coil->n_pri_r_o,
      coil->n_sec_h_b, coil->n_sec_h_t, coil->n_sec_r);
  (void)fprintf(fp,"\n\n");
  return;
  } /* end of function print_coil */



/**********************************************************************
   This function prints room data
**********************************************************************/
void print_room(FILE* fp, struct Grid_data *grid, int precision){

  char work_string[3][SI_STRING_SIZE];

  (void)fprintf(fp,
      "\n\nroom data:\n"
      "  radius = %sm\n"
      "  height = %sm\n"
      "  step size = %sm\n\n"
      "  discrete radius = %d\n"
      "  discrete height = %d\n",
      si_prefix(grid->room.r,precision,work_string[0]),
      si_prefix(grid->room.z,precision,work_string[1]),
      si_prefix(grid->room.dl,precision,work_string[2]),
      grid->room.n_r, grid->room.n_z);
  return;
  } /* end of function print_room */



/**********************************************************************
   This function prints interval range data
**********************************************************************/
void print_range(FILE* fp, struct Grid_data *grid, int precision){

  int i,j;

  /* Print range data */
  (void)fprintf(fp,
    "\n\n\nrange data:\n"
    "    row  #range       low    high        low    high ...\n");
  for(i=grid->room.n_z-1;i>=0;i--){
    (void)fprintf(fp,"\n%7d %7d   ",i,grid->range_h_count[i]);
    for(j=0;j<grid->range_h_count[i];j++){
      (void)fprintf(fp,"%7d %7d    ",
          grid->range_h_row[i][j].low,grid->range_h_row[i][j].high);
      }
    }

  (void)fprintf(fp,"\n\n");
  return;
  } /* end of function print_range */



/**********************************************************************
   This function prints a voltage matrix scaled by a factor of 
   10**precision 1<precision<=10.  Value= 10**precision indicated
   by *'s, value greater than this indicated by >'s and values less
   than 0 indicated by <'s.  The latter two cases indicate errors or
   numerical problems.
**********************************************************************/
void print_volt(FILE *fp, float **a, int r, int c, int precision){
  int i,j;
  float scale_factor=(float)pow(10.0,(double)precision);
  float at_limit_value=scale_factor-0.5f;
  float volt_scaled; /* scaled value of voltage */
  static char over_limit_string[]=">>>>>>>>>>"; /* string for too
                                                             large */
  static char under_limit_string[]="<<<<<<<<<<"; /* string for too
                                                             small */
  static char at_limit_string[]="**********"; /* string for maximum */
  static char format_string[]="%99.99f"; /* string to print V.
            (given specific values only to aid counting bytes) */

  /* Set the strings to the correct length for specified precision */
  over_limit_string[precision] = under_limit_string[precision] =
      at_limit_string[precision] = '\0';
  (void)sprintf(format_string," %%%d.0f",precision);

  /* Print header and meaning of symbols. */
  (void)fprintf(fp,
      "\n\n\nvoltage matrix:\n"
      "   %s=%f\n"
      "   %s= unexpectedly large number (error?)\n"
      "   %s= unexpectedly small number (error?)\n",
      at_limit_string,scale_factor,
      over_limit_string,
      under_limit_string);

  /* Print the voltage array */
  for(i=r-1;i>=0;i--){
    for(j=0;j<c;j++){
      volt_scaled = scale_factor*a[i][j];
      if(volt_scaled>scale_factor)
        (void)fprintf(fp," %s",over_limit_string);
      else if(volt_scaled>=at_limit_value)
        (void)fprintf(fp," %s",at_limit_string);
      else if(volt_scaled<0.0f)
        (void)fprintf(fp," %s",under_limit_string);
      else
        (void)fprintf(fp,format_string,volt_scaled);
      }
    (void)fprintf(fp,"\n");
    }
  (void)fprintf(fp,"\n\n");

  /* restore the characters in the >>>, <<<, and *** strings */
  over_limit_string[precision] = '>';
  under_limit_string[precision] = '<';
  at_limit_string[precision] = '*';
  
  return;
  } /* end of function print_volt */



/**********************************************************************
   This function prints the guard matrix.
**********************************************************************/
void print_guard(FILE *fp, GUARD_TYPE **a, int r, int c){

  int i,j;

  (void)fprintf(fp,"\n\n\nborder matrix:\n");
  for(i=r-1;i>=0;i--){
    for(j=0;j<c;j++){
      if(a[i][j]==(GUARD_TYPE)0)
        (void)fprintf(fp,"0");
      else if(a[i][j]==(GUARD_TYPE)1)
        (void)fprintf(fp,"1");
      else
        (void)fprintf(fp,"?");
      }
    (void)fprintf(fp,"\n");
    }
  return;
  } /* end of function print_guard */



/**********************************************************************
   This function converts a number to a string consisting of the
   number followed by a space and then the corresponding SI prefix.
   The number's integer portion is normalized to the range 1<|x|<999,
   except for the special case x=0.  
   
   The precision must be >0 and is the total number of digits in the
   number.
   
   Work_string must be a pointer to a string of at least
   precision+4+EXPONENT_WIDTH+1 characters, including trailing null.

   If the number is out of the range of SI prefixes, it is converted
   to e-format and the string is returned without any attempt to
   convert it to SI-form.
**********************************************************************/
char *si_prefix(double x, int precision, char *work_string){

  static char error_string[]="???";
  const static char pos_prefix[]="?kMGTPE"; /* SI prefixes 3,6,... */
  const static char neg_prefix[]="?munpfa"; /* SI prefixes -3,-6,... */

  char *loc; /* locator in work string */
  char *loc_i; /* loc of start of integer portion */
  char *loc_dot; /* loc of '.' */
  char *loc_e; /* loc of 'e' */
  char prefix_char; /* character to indicate scaling */
  int i; /* index */
  int exponent; /* integer form (not character) of exponent */
  int decimal_shift; /* amount to shift decimal point */
  int prefix_index; /* index to access char of one of the
                                           prefix strings */

  if(precision<=0)return error_string; /* invalid precision */

  /* Convert number to e-format string.  There should never be an
     error here unless some extreme values are used for special
     purposes and do not convert.  The specified format will
     generate the exact number of digits required, except
     possibly some trailing 0s that result from shifting the
     decimal point. */
  if(sprintf(work_string,"%*.*e",precision+4+EXPONENT_WIDTH,
      precision-1,x)==EOF)
    return error_string;


  /* The string is now scanned to find its parts:

       leading_blanks    integer    .    fraction    'e'    exponent
                     opt             opt         opt
  */

  /* Step past leading blanks (0 to 2 depending on +/- and p>0 / p=0 */
  for(loc=work_string;*loc==' ';loc++);

  loc_i=loc; /* save ptr to integer part (-D or D) */

  /* If out of range, just return the string with blank at end.
     The function would operate correctly without this check because
     the out-of-range condition would be detected later. */
  if(x>0&&(x<1E-18 || x>=1E21) || x<0&&(-x<1E-18 || -x>=1E21))
    return loc_i;


  /* Step past fractional portion until reach 'e'.  Note that a '.'
     may or may not appear depending on specified precision (0 -> no
     decimal point).  Consequently, the '.' is *NOT* looked for next */
  for(;*loc!='e';loc++);
  loc_e=loc; /* save ptr to 'e' exponent indicator */

  /* Find decimal point (or 'e' if none) */
  for(loc=work_string;loc<loc_e&&*loc!='.';loc++);
  loc_dot=loc; /* save ptr to '.' or 'e' */

  /* Convert exponent to an integer.  This conversion should never
     fail.   The space after the %d is to handle the case of the
     number being too large or small for SI units.  In this case,
     the number is returned in exponential notation with a space
     following the exponent. */
  (void)sscanf(loc_e+1,"%d ",&exponent);

  /* % is machine-dependent for negative numbers, so always use
     non-negatives. */
  /* The returns handles a few remaining cases where the number would
     have been convertable to SI-prefix form, but the precision
     that was specified caused rounding up to a non-existent prefix */
  if(exponent>=0){
    prefix_index = exponent/3;
    if(prefix_index>6) return loc_i;
    decimal_shift =exponent%3;
    prefix_char=pos_prefix[prefix_index];
    }
  else{
    prefix_index = (-exponent+2)/3;
    if(prefix_index>6) return loc_i;
    decimal_shift = (999+exponent)%3; /* 999 is a multiple m of 3
                                        such that m + exponent > 0 */
    prefix_char=neg_prefix[prefix_index];
    }


  /* Re-format the string with the prefix replacing the exponent
     and the decimal point shifted */

  /* Overwrite decimal point with digits until run out of digits.
     This will occur if the required precision is < decimal_shift */ 
  for(loc=loc_dot,i=0;loc<loc_e-1&&i<decimal_shift;loc++,i++){
    loc[0]=loc[1];
    }

  /* Either add a '.' and leave the rest of the fraction as is
     or add 0's */
  if(i<decimal_shift){
    /* Fill remaining places with 0s */
    for(               ;          i<decimal_shift;loc++,i++){
      loc[0]='0';   
      }
    }
  else{
    *loc++='.';
    loc = loc_e; /* point to position after fraction */
    }

  *loc++=' '; /* add space after number */

  /* Add the scaling prefix unless |number|<1000.  This will occur if
     the prefix index is 0. */
  if(prefix_index!=0){
    *loc++=prefix_char;
    }

  *loc='\0'; /* terminate the string */

  return loc_i; /* return at start of integer part (may be - sign) */
  } /* end of function si_prefix */



#ifdef DEBUG_freq
/**********************************************************************
   This function compares the voltage arrays and identfies any elements
   which are in disagreement.
**********************************************************************/
int compare_v(char* filename, char* tag, struct Grid_data *grid1,
    struct Grid_data *grid2){
  int i;
  int first_error=0;
  float err;
  FILE* fp;

  /* Open the diagostic file and append data, unless stderr was
     requested. */
  if(filename[0]=='-'&&filename[1]=='\0'){
    fp=stderr;
    }
  else{
    if((fp=fopen(filename,"a"))==(FILE*)NULL){
      (void)fprintf(stderr,"error opening diagnostic file\n");
      return 1;
      }
    }

  /* Compare element by element and print those that are further
     than TOL_V apart */
  (void)fprintf(fp,
      "%s -- "
      "comparing voltage matrices...",tag);
  for(i=0;i<grid1->room.n_r*grid1->room.n_z;i++){
    if((err=(float)fabs((double)(grid1->v[i]-grid2->v[i])))>
        TOL_V*MAX(grid1->v[i],grid2->v[i])){
      if(first_error==0){ /* if this is first error */
        first_error=1; /* no longer first error */
        (void)fprintf(fp,
            "\n"
            "List of points in error:\n\n"
            "   row    column      V from v      V from vv "   
                "     rel error\n\n");
        } /* end of 1st error */
      (void)fprintf(fp,"%5d     %5d   %12.5g   %12.5g   %12.5g\n",
          i/grid1->room.n_r,i%grid1->room.n_r,grid1->v[i],grid2->v[i],
          err);
      } /* end of if disagree */
    } /* end of loop comparing element by element */
      if(first_error==0)(void)fprintf(fp,"  matrices agree\n");

  if(fp!=stderr)(void)fclose(fp); /* close so can look at */
  return 0;
  } /* end of function compare_v */
/*********************************************************************/
#endif /* DEBUG_freq */
